@@ -3,8 +3,6 @@ from io import BytesIO
 import logging
 import os
 import stat
-from typing import List
-from unittest import SkipTest
 import uuid
 
 import git
@@ -25,14 +23,15 @@ from git.exc import (
     BadName
 )
 from git.objects.base import IndexObject, Object
-from git.objects.util import Traversable
+from git.objects.util import TraversableIterableObj
+
 from git.util import (
-    Iterable,
     join_path_native,
     to_native_path_linux,
     RemoteProgress,
     rmtree,
-    unbare_repo
+    unbare_repo,
+    IterableList
 )
 from git.util import HIDE_WINDOWS_KNOWN_ERRORS
 
@@ -47,6 +46,20 @@ from .util import (
 )
 
 
+# typing ----------------------------------------------------------------------
+from typing import Callable, Dict, Mapping, Sequence, TYPE_CHECKING, cast
+from typing import Any, Iterator, Union
+
+from git.types import Commit_ish, Literal, PathLike, TBD
+
+if TYPE_CHECKING:
+    from git.index import IndexFile
+    from git.repo import Repo
+    from git.refs import Head
+
+
+# -----------------------------------------------------------------------------
+
 __all__ = ["Submodule", "UpdateProgress"]
 
 
@@ -59,7 +72,7 @@ class UpdateProgress(RemoteProgress):
     """Class providing detailed progress information to the caller who should
     derive from it and implement the ``update(...)`` message"""
     CLONE, FETCH, UPDWKTREE = [1 << x for x in range(RemoteProgress._num_op_codes, RemoteProgress._num_op_codes + 3)]
-    _num_op_codes = RemoteProgress._num_op_codes + 3
+    _num_op_codes: int = RemoteProgress._num_op_codes + 3
 
     __slots__ = ()
 
@@ -74,7 +87,7 @@ UPDWKTREE = UpdateProgress.UPDWKTREE
 # IndexObject comes via util module, its a 'hacky' fix thanks to pythons import
 # mechanism which cause plenty of trouble of the only reason for packages and
 # modules is refactoring - subpackages shouldn't depend on parent packages
-class Submodule(IndexObject, Iterable, Traversable):
+class Submodule(IndexObject, TraversableIterableObj):
 
     """Implements access to a git submodule. They are special in that their sha
     represents a commit in the submodule's repository which is to be checked out
@@ -91,12 +104,19 @@ class Submodule(IndexObject, Iterable, Traversable):
     k_default_mode = stat.S_IFDIR | stat.S_IFLNK        # submodules are directories with link-status
 
     # this is a bogus type for base class compatibility
-    type = 'submodule'
+    type: Literal['submodule'] = 'submodule'  # type: ignore
 
     __slots__ = ('_parent_commit', '_url', '_branch_path', '_name', '__weakref__')
     _cache_attrs = ('path', '_url', '_branch_path')
 
-    def __init__(self, repo, binsha, mode=None, path=None, name=None, parent_commit=None, url=None, branch_path=None):
+    def __init__(self, repo: 'Repo', binsha: bytes,
+                 mode: Union[int, None] = None,
+                 path: Union[PathLike, None] = None,
+                 name: Union[str, None] = None,
+                 parent_commit: Union[Commit_ish, None] = None,
+                 url: Union[str, None] = None,
+                 branch_path: Union[PathLike, None] = None
+                 ) -> None:
         """Initialize this instance with its attributes. We only document the ones
         that differ from ``IndexObject``
 
@@ -111,20 +131,21 @@ class Submodule(IndexObject, Iterable, Traversable):
         if url is not None:
             self._url = url
         if branch_path is not None:
-            assert isinstance(branch_path, str)
+            # assert isinstance(branch_path, str)
             self._branch_path = branch_path
         if name is not None:
             self._name = name
 
-    def _set_cache_(self, attr):
+    def _set_cache_(self, attr: str) -> None:
         if attr in ('path', '_url', '_branch_path'):
-            reader = self.config_reader()
+            reader: SectionConstraint = self.config_reader()
             # default submodule values
             try:
                 self.path = reader.get('path')
             except cp.NoSectionError as e:
-                raise ValueError("This submodule instance does not exist anymore in '%s' file"
-                                 % osp.join(self.repo.working_tree_dir, '.gitmodules')) from e
+                if self.repo.working_tree_dir is not None:
+                    raise ValueError("This submodule instance does not exist anymore in '%s' file"
+                                     % osp.join(self.repo.working_tree_dir, '.gitmodules')) from e
             # end
             self._url = reader.get('url')
             # git-python extension values - optional
@@ -136,42 +157,44 @@ class Submodule(IndexObject, Iterable, Traversable):
         # END handle attribute name
 
     @classmethod
-    def _get_intermediate_items(cls, item: 'Submodule') -> List['Submodule']:  # type: ignore
+    def _get_intermediate_items(cls, item: 'Submodule') -> IterableList['Submodule']:
         """:return: all the submodules of our module repository"""
         try:
             return cls.list_items(item.module())
         except InvalidGitRepositoryError:
-            return []
+            return IterableList('')
         # END handle intermediate items
 
     @classmethod
-    def _need_gitfile_submodules(cls, git):
+    def _need_gitfile_submodules(cls, git: Git) -> bool:
         return git.version_info[:3] >= (1, 7, 5)
 
-    def __eq__(self, other):
+    def __eq__(self, other: Any) -> bool:
         """Compare with another submodule"""
         # we may only compare by name as this should be the ID they are hashed with
         # Otherwise this type wouldn't be hashable
         # return self.path == other.path and self.url == other.url and super(Submodule, self).__eq__(other)
         return self._name == other._name
 
-    def __ne__(self, other):
+    def __ne__(self, other: object) -> bool:
         """Compare with another submodule for inequality"""
         return not (self == other)
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         """Hash this instance using its logical id, not the sha"""
         return hash(self._name)
 
-    def __str__(self):
+    def __str__(self) -> str:
         return self._name
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "git.%s(name=%s, path=%s, url=%s, branch_path=%s)"\
                % (type(self).__name__, self._name, self.path, self.url, self.branch_path)
 
     @classmethod
-    def _config_parser(cls, repo, parent_commit, read_only):
+    def _config_parser(cls, repo: 'Repo',
+                       parent_commit: Union[Commit_ish, None],
+                       read_only: bool) -> SubmoduleConfigParser:
         """:return: Config Parser constrained to our submodule in read or write mode
         :raise IOError: If the .gitmodules file cannot be found, either locally or in the repository
             at the given parent commit. Otherwise the exception would be delayed until the first
@@ -184,8 +207,8 @@ class Submodule(IndexObject, Iterable, Traversable):
                 # We are most likely in an empty repository, so the HEAD doesn't point to a valid ref
                 pass
         # end handle parent_commit
-
-        if not repo.bare and parent_matches_head:
+        fp_module: Union[str, BytesIO]
+        if not repo.bare and parent_matches_head and repo.working_tree_dir:
             fp_module = osp.join(repo.working_tree_dir, cls.k_modules_file)
         else:
             assert parent_commit is not None, "need valid parent_commit in bare repositories"
@@ -203,7 +226,7 @@ class Submodule(IndexObject, Iterable, Traversable):
 
         return SubmoduleConfigParser(fp_module, read_only=read_only)
 
-    def _clear_cache(self):
+    def _clear_cache(self) -> None:
         # clear the possibly changed values
         for name in self._cache_attrs:
             try:
@@ -214,16 +237,16 @@ class Submodule(IndexObject, Iterable, Traversable):
         # END for each name to delete
 
     @classmethod
-    def _sio_modules(cls, parent_commit):
+    def _sio_modules(cls, parent_commit: Commit_ish) -> BytesIO:
         """:return: Configuration file as BytesIO - we only access it through the respective blob's data"""
         sio = BytesIO(parent_commit.tree[cls.k_modules_file].data_stream.read())
         sio.name = cls.k_modules_file
         return sio
 
-    def _config_parser_constrained(self, read_only):
+    def _config_parser_constrained(self, read_only: bool) -> SectionConstraint:
         """:return: Config Parser constrained to our submodule in read or write mode"""
         try:
-            pc = self.parent_commit
+            pc: Union['Commit_ish', None] = self.parent_commit
         except ValueError:
             pc = None
         # end handle empty parent repository
@@ -232,18 +255,20 @@ class Submodule(IndexObject, Iterable, Traversable):
         return SectionConstraint(parser, sm_section(self.name))
 
     @classmethod
-    def _module_abspath(cls, parent_repo, path, name):
+    def _module_abspath(cls, parent_repo: 'Repo', path: PathLike, name: str) -> PathLike:
         if cls._need_gitfile_submodules(parent_repo.git):
             return osp.join(parent_repo.git_dir, 'modules', name)
-        return osp.join(parent_repo.working_tree_dir, path)
+        if parent_repo.working_tree_dir:
+            return osp.join(parent_repo.working_tree_dir, path)
+        raise NotADirectoryError()
         # end
 
     @classmethod
-    def _clone_repo(cls, repo, url, path, name, **kwargs):
+    def _clone_repo(cls, repo: 'Repo', url: str, path: PathLike, name: str, **kwargs: Any) -> 'Repo':
         """:return: Repo instance of newly cloned repository
         :param repo: our parent repository
         :param url: url to clone from
-        :param path: repository-relative path to the submodule checkout location
+        :param path: repository - relative path to the submodule checkout location
         :param name: canonical of the submodule
         :param kwrags: additinoal arguments given to git.clone"""
         module_abspath = cls._module_abspath(repo, path, name)
@@ -253,7 +278,7 @@ class Submodule(IndexObject, Iterable, Traversable):
             module_abspath_dir = osp.dirname(module_abspath)
             if not osp.isdir(module_abspath_dir):
                 os.makedirs(module_abspath_dir)
-            module_checkout_path = osp.join(repo.working_tree_dir, path)
+            module_checkout_path = osp.join(str(repo.working_tree_dir), path)
         # end
 
         clone = git.Repo.clone_from(url, module_checkout_path, **kwargs)
@@ -263,15 +288,15 @@ class Submodule(IndexObject, Iterable, Traversable):
         return clone
 
     @classmethod
-    def _to_relative_path(cls, parent_repo, path):
-        """:return: a path guaranteed  to be relative to the given parent-repository
+    def _to_relative_path(cls, parent_repo: 'Repo', path: PathLike) -> PathLike:
+        """:return: a path guaranteed  to be relative to the given parent - repository
         :raise ValueError: if path is not contained in the parent repository's working tree"""
         path = to_native_path_linux(path)
         if path.endswith('/'):
             path = path[:-1]
         # END handle trailing slash
 
-        if osp.isabs(path):
+        if osp.isabs(path) and parent_repo.working_tree_dir:
             working_tree_linux = to_native_path_linux(parent_repo.working_tree_dir)
             if not path.startswith(working_tree_linux):
                 raise ValueError("Submodule checkout path '%s' needs to be within the parents repository at '%s'"
@@ -285,12 +310,12 @@ class Submodule(IndexObject, Iterable, Traversable):
         return path
 
     @classmethod
-    def _write_git_file_and_module_config(cls, working_tree_dir, module_abspath):
-        """Writes a .git file containing a (preferably) relative path to the actual git module repository.
+    def _write_git_file_and_module_config(cls, working_tree_dir: PathLike, module_abspath: PathLike) -> None:
+        """Writes a .git file containing a(preferably) relative path to the actual git module repository.
         It is an error if the module_abspath cannot be made into a relative path, relative to the working_tree_dir
         :note: will overwrite existing files !
         :note: as we rewrite both the git file as well as the module configuration, we might fail on the configuration
-            and will not roll back changes done to the git file. This should be a non-issue, but may easily be fixed
+            and will not roll back changes done to the git file. This should be a non - issue, but may easily be fixed
             if it becomes one
         :param working_tree_dir: directory to write the .git file into
         :param module_abspath: absolute path to the bare repository
@@ -311,7 +336,10 @@ class Submodule(IndexObject, Iterable, Traversable):
     #{ Edit Interface
 
     @classmethod
-    def add(cls, repo, name, path, url=None, branch=None, no_checkout=False, depth=None, env=None):
+    def add(cls, repo: 'Repo', name: str, path: PathLike, url: Union[str, None] = None,
+            branch: Union[str, None] = None, no_checkout: bool = False, depth: Union[int, None] = None,
+            env: Union[Mapping[str, str], None] = None, clone_multi_options: Union[Sequence[TBD], None] = None
+            ) -> 'Submodule':
         """Add a new submodule to the given repository. This will alter the index
         as well as the .gitmodules file, but will not create a new commit.
         If the submodule already exists, no matter if the configuration differs
@@ -344,9 +372,12 @@ class Submodule(IndexObject, Iterable, Traversable):
             and is defined in `os.environ`, value from `os.environ` will be used.
             If you want to unset some variable, consider providing empty string
             as its value.
+        :param clone_multi_options: A list of Clone options. Please see ``git.repo.base.Repo.clone``
+            for details.
         :return: The newly created submodule instance
         :note: works atomically, such that no change will be done if the repository
             update fails for instance"""
+
         if repo.bare:
             raise InvalidGitRepositoryError("Cannot add submodules to bare repositories")
         # END handle bare repos
@@ -364,7 +395,7 @@ class Submodule(IndexObject, Iterable, Traversable):
         if sm.exists():
             # reretrieve submodule from tree
             try:
-                sm = repo.head.commit.tree[path]
+                sm = repo.head.commit.tree[str(path)]
                 sm._name = name
                 return sm
             except KeyError:
@@ -387,12 +418,14 @@ class Submodule(IndexObject, Iterable, Traversable):
             # END check url
         # END verify urls match
 
-        mrepo = None
+        mrepo: Union[Repo, None] = None
+
         if url is None:
             if not has_module:
-                raise ValueError("A URL was not given and existing repository did not exsit at %s" % path)
+                raise ValueError("A URL was not given and a repository did not exist at %s" % path)
             # END check url
             mrepo = sm.module()
+            # assert isinstance(mrepo, git.Repo)
             urls = [r.url for r in mrepo.remotes]
             if not urls:
                 raise ValueError("Didn't find any remote url in repository at %s" % sm.abspath)
@@ -400,7 +433,7 @@ class Submodule(IndexObject, Iterable, Traversable):
             url = urls[0]
         else:
             # clone new repo
-            kwargs = {'n': no_checkout}
+            kwargs: Dict[str, Union[bool, int, str, Sequence[TBD]]] = {'n': no_checkout}
             if not branch_is_default:
                 kwargs['b'] = br.name
             # END setup checkout-branch
@@ -410,6 +443,8 @@ class Submodule(IndexObject, Iterable, Traversable):
                     kwargs['depth'] = depth
                 else:
                     raise ValueError("depth should be an integer")
+            if clone_multi_options:
+                kwargs['multi_options'] = clone_multi_options
 
             # _clone_repo(cls, repo, url, path, name, **kwargs):
             mrepo = cls._clone_repo(repo, url, path, name, env=env, **kwargs)
@@ -422,6 +457,8 @@ class Submodule(IndexObject, Iterable, Traversable):
         # otherwise there is a '-' character in front of the submodule listing
         #  a38efa84daef914e4de58d1905a500d8d14aaf45 mymodule (v0.9.0-1-ga38efa8)
         # -a38efa84daef914e4de58d1905a500d8d14aaf45 submodules/intermediate/one
+        writer: Union[GitConfigParser, SectionConstraint]
+
         with sm.repo.config_writer() as writer:
             writer.set_value(sm_section(name), 'url', url)
 
@@ -438,13 +475,16 @@ class Submodule(IndexObject, Iterable, Traversable):
                 sm._branch_path = br.path
 
         # we deliberately assume that our head matches our index !
-        sm.binsha = mrepo.head.commit.binsha
+        if mrepo:
+            sm.binsha = mrepo.head.commit.binsha
         index.add([sm], write=True)
 
         return sm
 
-    def update(self, recursive=False, init=True, to_latest_revision=False, progress=None, dry_run=False,
-               force=False, keep_going=False, env=None):
+    def update(self, recursive: bool = False, init: bool = True, to_latest_revision: bool = False,
+               progress: Union['UpdateProgress', None] = None, dry_run: bool = False,
+               force: bool = False, keep_going: bool = False, env: Union[Mapping[str, str], None] = None,
+               clone_multi_options: Union[Sequence[TBD], None] = None) -> 'Submodule':
         """Update the repository of this submodule to point to the checkout
         we point at with the binsha of this instance.
 
@@ -458,7 +498,7 @@ class Submodule(IndexObject, Iterable, Traversable):
             was specified for this submodule and the branch existed remotely
         :param progress: UpdateProgress instance or None if no progress should be shown
         :param dry_run: if True, the operation will only be simulated, but not performed.
-            All performed operations are read-only
+            All performed operations are read - only
         :param force:
             If True, we may reset heads even if the repository in question is dirty. Additinoally we will be allowed
             to set a tracking branch which is ahead of its remote branch back into the past or the location of the
@@ -466,7 +506,7 @@ class Submodule(IndexObject, Iterable, Traversable):
             If False, local tracking branches that are in the future of their respective remote branches will simply
             not be moved.
         :param keep_going: if True, we will ignore but log all errors, and keep going recursively.
-            Unless dry_run is set as well, keep_going could cause subsequent/inherited errors you wouldn't see
+            Unless dry_run is set as well, keep_going could cause subsequent / inherited errors you wouldn't see
             otherwise.
             In conjunction with dry_run, it can be useful to anticipate all errors when updating submodules
         :param env: Optional dictionary containing the desired environment variables.
@@ -475,6 +515,8 @@ class Submodule(IndexObject, Iterable, Traversable):
             and is defined in `os.environ`, value from `os.environ` will be used.
             If you want to unset some variable, consider providing empty string
             as its value.
+        :param clone_multi_options:  list of Clone options. Please see ``git.repo.base.Repo.clone``
+            for details. Only take effect with `init` option.
         :note: does nothing in bare repositories
         :note: method is definitely not atomic if recurisve is True
         :return: self"""
@@ -521,6 +563,7 @@ class Submodule(IndexObject, Iterable, Traversable):
                     progress.update(op, i, len_rmts, prefix + "Done fetching remote of submodule %r" % self.name)
                 # END fetch new data
             except InvalidGitRepositoryError:
+                mrepo = None
                 if not init:
                     return self
                 # END early abort if init is not allowed
@@ -541,13 +584,15 @@ class Submodule(IndexObject, Iterable, Traversable):
                 progress.update(BEGIN | CLONE, 0, 1, prefix + "Cloning url '%s' to '%s' in submodule %r" %
                                 (self.url, checkout_module_abspath, self.name))
                 if not dry_run:
-                    mrepo = self._clone_repo(self.repo, self.url, self.path, self.name, n=True, env=env)
+                    mrepo = self._clone_repo(self.repo, self.url, self.path, self.name, n=True, env=env,
+                                             multi_options=clone_multi_options)
                 # END handle dry-run
                 progress.update(END | CLONE, 0, 1, prefix + "Done cloning to %s" % checkout_module_abspath)
 
                 if not dry_run:
                     # see whether we have a valid branch to checkout
                     try:
+                        mrepo = cast('Repo', mrepo)
                         # find  a remote which has our branch - we try to be flexible
                         remote_branch = find_first_remote_branch(mrepo.remotes, self.branch_name)
                         local_branch = mkhead(mrepo, self.branch_path)
@@ -559,7 +604,7 @@ class Submodule(IndexObject, Iterable, Traversable):
 
                         # make sure HEAD is not detached
                         mrepo.head.set_reference(local_branch, logmsg="submodule: attaching head to %s" % local_branch)
-                        mrepo.head.ref.set_tracking_branch(remote_branch)
+                        mrepo.head.reference.set_tracking_branch(remote_branch)
                     except (IndexError, InvalidGitRepositoryError):
                         log.warning("Failed to checkout tracking branch %s", self.branch_path)
                     # END handle tracking branch
@@ -585,13 +630,14 @@ class Submodule(IndexObject, Iterable, Traversable):
             if mrepo is not None and to_latest_revision:
                 msg_base = "Cannot update to latest revision in repository at %r as " % mrepo.working_dir
                 if not is_detached:
-                    rref = mrepo.head.ref.tracking_branch()
+                    rref = mrepo.head.reference.tracking_branch()
                     if rref is not None:
                         rcommit = rref.commit
                         binsha = rcommit.binsha
                         hexsha = rcommit.hexsha
                     else:
-                        log.error("%s a tracking branch was not set for local branch '%s'", msg_base, mrepo.head.ref)
+                        log.error("%s a tracking branch was not set for local branch '%s'",
+                                  msg_base, mrepo.head.reference)
                     # END handle remote ref
                 else:
                     log.error("%s there was no local tracking branch", msg_base)
@@ -608,7 +654,7 @@ class Submodule(IndexObject, Iterable, Traversable):
                 may_reset = True
                 if mrepo.head.commit.binsha != self.NULL_BIN_SHA:
                     base_commit = mrepo.merge_base(mrepo.head.commit, hexsha)
-                    if len(base_commit) == 0 or base_commit[0].hexsha == hexsha:
+                    if len(base_commit) == 0 or (base_commit[0] is not None and base_commit[0].hexsha == hexsha):
                         if force:
                             msg = "Will force checkout or reset on local branch that is possibly in the future of"
                             msg += "the commit it will be checked out to, effectively 'forgetting' new commits"
@@ -666,15 +712,15 @@ class Submodule(IndexObject, Iterable, Traversable):
         return self
 
     @unbare_repo
-    def move(self, module_path, configuration=True, module=True):
+    def move(self, module_path: PathLike, configuration: bool = True, module: bool = True) -> 'Submodule':
         """Move the submodule to a another module path. This involves physically moving
         the repository at our current path, changing the configuration, as well as
         adjusting our index entry accordingly.
 
         :param module_path: the path to which to move our module in the parent repostory's working tree,
-            given as repository-relative or absolute path. Intermediate directories will be created
+            given as repository - relative or absolute path. Intermediate directories will be created
             accordingly. If the path already exists, it must be empty.
-            Trailing (back)slashes are removed automatically
+            Trailing(back)slashes are removed automatically
         :param configuration: if True, the configuration will be adjusted to let
             the submodule point to the given path.
         :param module: if True, the repository managed by this submodule
@@ -683,7 +729,7 @@ class Submodule(IndexObject, Iterable, Traversable):
         :return: self
         :raise ValueError: if the module path existed and was not empty, or was a file
         :note: Currently the method is not atomic, and it could leave the repository
-            in an inconsistent state if a sub-step fails for some reason
+            in an inconsistent state if a sub - step fails for some reason
         """
         if module + configuration < 1:
             raise ValueError("You must specify to move at least the module or the configuration of the submodule")
@@ -696,7 +742,7 @@ class Submodule(IndexObject, Iterable, Traversable):
             return self
         # END handle no change
 
-        module_checkout_abspath = join_path_native(self.repo.working_tree_dir, module_checkout_path)
+        module_checkout_abspath = join_path_native(str(self.repo.working_tree_dir), module_checkout_path)
         if osp.isfile(module_checkout_abspath):
             raise ValueError("Cannot move repository onto a file: %s" % module_checkout_abspath)
         # END handle target files
@@ -775,21 +821,22 @@ class Submodule(IndexObject, Iterable, Traversable):
         return self
 
     @unbare_repo
-    def remove(self, module=True, force=False, configuration=True, dry_run=False):
+    def remove(self, module: bool = True, force: bool = False,
+               configuration: bool = True, dry_run: bool = False) -> 'Submodule':
         """Remove this submodule from the repository. This will remove our entry
-        from the .gitmodules file and the entry in the .git/config file.
+        from the .gitmodules file and the entry in the .git / config file.
 
         :param module: If True, the module checkout we point to will be deleted
             as well. If the module is currently on a commit which is not part
             of any branch in the remote, if the currently checked out branch
             working tree, or untracked files,
-            is ahead of its tracking branch,  if you have modifications in the
+            is ahead of its tracking branch, if you have modifications in the
             In case the removal of the repository fails for these reasons, the
             submodule status will not have been altered.
-            If this submodule has child-modules on its own, these will be deleted
+            If this submodule has child - modules on its own, these will be deleted
             prior to touching the own module.
         :param force: Enforces the deletion of the module even though it contains
-            modifications. This basically enforces a brute-force file system based
+            modifications. This basically enforces a brute - force file system based
             deletion.
         :param configuration: if True, the submodule is deleted from the configuration,
             otherwise it isn't. Although this should be enabled most of the times,
@@ -829,7 +876,7 @@ class Submodule(IndexObject, Iterable, Traversable):
                 # TODO: If we run into permission problems, we have a highly inconsistent
                 # state. Delete the .git folders last, start with the submodules first
                 mp = self.abspath
-                method = None
+                method: Union[None, Callable[[PathLike], None]] = None
                 if osp.islink(mp):
                     method = os.remove
                 elif osp.isdir(mp):
@@ -882,9 +929,10 @@ class Submodule(IndexObject, Iterable, Traversable):
                     import gc
                     gc.collect()
                     try:
-                        rmtree(wtd)
+                        rmtree(str(wtd))
                     except Exception as ex:
                         if HIDE_WINDOWS_KNOWN_ERRORS:
+                            from unittest import SkipTest
                             raise SkipTest("FIXME: fails with: PermissionError\n  {}".format(ex)) from ex
                         raise
                 # END delete tree if possible
@@ -896,7 +944,8 @@ class Submodule(IndexObject, Iterable, Traversable):
                     rmtree(git_dir)
                 except Exception as ex:
                     if HIDE_WINDOWS_KNOWN_ERRORS:
-                        raise SkipTest("FIXME: fails with: PermissionError\n  %s", ex) from ex
+                        from unittest import SkipTest
+                        raise SkipTest(f"FIXME: fails with: PermissionError\n  {ex}") from ex
                     else:
                         raise
             # end handle separate bare repository
@@ -920,16 +969,17 @@ class Submodule(IndexObject, Iterable, Traversable):
 
             # now git config - need the config intact, otherwise we can't query
             # information anymore
-            with self.repo.config_writer() as writer:
-                writer.remove_section(sm_section(self.name))
 
-            with self.config_writer() as writer:
-                writer.remove_section()
+            with self.repo.config_writer() as gcp_writer:
+                gcp_writer.remove_section(sm_section(self.name))
+
+            with self.config_writer() as sc_writer:
+                sc_writer.remove_section()
         # END delete configuration
 
         return self
 
-    def set_parent_commit(self, commit, check=True):
+    def set_parent_commit(self, commit: Union[Commit_ish, None], check: bool = True) -> 'Submodule':
         """Set this instance to use the given commit whose tree is supposed to
         contain the .gitmodules blob.
 
@@ -968,7 +1018,7 @@ class Submodule(IndexObject, Iterable, Traversable):
         # If check is False, we might see a parent-commit that doesn't even contain the submodule anymore.
         # in that case, mark our sha as being NULL
         try:
-            self.binsha = pctree[self.path].binsha
+            self.binsha = pctree[str(self.path)].binsha
         except KeyError:
             self.binsha = self.NULL_BIN_SHA
         # end
@@ -977,7 +1027,8 @@ class Submodule(IndexObject, Iterable, Traversable):
         return self
 
     @unbare_repo
-    def config_writer(self, index=None, write=True):
+    def config_writer(self, index: Union['IndexFile', None] = None, write: bool = True
+                      ) -> SectionConstraint['SubmoduleConfigParser']:
         """:return: a config writer instance allowing you to read and write the data
             belonging to this submodule into the .gitmodules file.
 
@@ -998,13 +1049,13 @@ class Submodule(IndexObject, Iterable, Traversable):
         return writer
 
     @unbare_repo
-    def rename(self, new_name):
+    def rename(self, new_name: str) -> 'Submodule':
         """Rename this submodule
         :note: This method takes care of renaming the submodule in various places, such as
 
-            * $parent_git_dir/config
-            * $working_tree_dir/.gitmodules
-            * (git >=v1.8.0: move submodule repository to new name)
+            * $parent_git_dir / config
+            * $working_tree_dir / .gitmodules
+            * (git >= v1.8.0: move submodule repository to new name)
 
         As .gitmodules will be changed, you would need to make a commit afterwards. The changed .gitmodules file
         will already be added to the index
@@ -1033,13 +1084,14 @@ class Submodule(IndexObject, Iterable, Traversable):
             destination_module_abspath = self._module_abspath(self.repo, self.path, new_name)
             source_dir = mod.git_dir
             # Let's be sure the submodule name is not so obviously tied to a directory
-            if destination_module_abspath.startswith(mod.git_dir):
+            if str(destination_module_abspath).startswith(str(mod.git_dir)):
                 tmp_dir = self._module_abspath(self.repo, self.path, str(uuid.uuid4()))
                 os.renames(source_dir, tmp_dir)
                 source_dir = tmp_dir
             # end handle self-containment
             os.renames(source_dir, destination_module_abspath)
-            self._write_git_file_and_module_config(mod.working_tree_dir, destination_module_abspath)
+            if mod.working_tree_dir:
+                self._write_git_file_and_module_config(mod.working_tree_dir, destination_module_abspath)
         # end move separate git repository
 
         return self
@@ -1049,7 +1101,7 @@ class Submodule(IndexObject, Iterable, Traversable):
     #{ Query Interface
 
     @unbare_repo
-    def module(self):
+    def module(self) -> 'Repo':
         """:return: Repo instance initialized from the repository at our submodule path
         :raise InvalidGitRepositoryError: if a repository was not available. This could
             also mean that it was not yet initialized"""
@@ -1066,7 +1118,7 @@ class Submodule(IndexObject, Iterable, Traversable):
             raise InvalidGitRepositoryError("Repository at %r was not yet checked out" % module_checkout_abspath)
         # END handle exceptions
 
-    def module_exists(self):
+    def module_exists(self) -> bool:
         """:return: True if our module exists and is a valid git repository. See module() method"""
         try:
             self.module()
@@ -1075,10 +1127,10 @@ class Submodule(IndexObject, Iterable, Traversable):
             return False
         # END handle exception
 
-    def exists(self):
+    def exists(self) -> bool:
         """
         :return: True if the submodule exists, False otherwise. Please note that
-            a submodule may exist (in the .gitmodules file) even though its module
+            a submodule may exist ( in the .gitmodules file) even though its module
             doesn't exist on disk"""
         # keep attributes for later, and restore them if we have no valid data
         # this way we do not actually alter the state of the object
@@ -1110,32 +1162,32 @@ class Submodule(IndexObject, Iterable, Traversable):
         # END handle object state consistency
 
     @property
-    def branch(self):
+    def branch(self) -> 'Head':
         """:return: The branch instance that we are to checkout
         :raise InvalidGitRepositoryError: if our module is not yet checked out"""
         return mkhead(self.module(), self._branch_path)
 
     @property
-    def branch_path(self):
+    def branch_path(self) -> PathLike:
         """
-        :return: full (relative) path as string to the branch we would checkout
+        :return: full(relative) path as string to the branch we would checkout
             from the remote and track"""
         return self._branch_path
 
     @property
-    def branch_name(self):
+    def branch_name(self) -> str:
         """:return: the name of the branch, which is the shortest possible branch name"""
         # use an instance method, for this we create a temporary Head instance
         # which uses a repository that is available at least ( it makes no difference )
         return git.Head(self.repo, self._branch_path).name
 
     @property
-    def url(self):
-        """:return: The url to the repository which our module-repository refers to"""
+    def url(self) -> str:
+        """:return: The url to the repository which our module - repository refers to"""
         return self._url
 
     @property
-    def parent_commit(self):
+    def parent_commit(self) -> 'Commit_ish':
         """:return: Commit instance with the tree containing the .gitmodules file
         :note: will always point to the current head's commit if it was not set explicitly"""
         if self._parent_commit is None:
@@ -1143,17 +1195,17 @@ class Submodule(IndexObject, Iterable, Traversable):
         return self._parent_commit
 
     @property
-    def name(self):
+    def name(self) -> str:
         """:return: The name of this submodule. It is used to identify it within the
             .gitmodules file.
         :note: by default, the name is the path at which to find the submodule, but
-            in git-python it should be a unique identifier similar to the identifiers
+            in git - python it should be a unique identifier similar to the identifiers
             used for remotes, which allows to change the path of the submodule
             easily
         """
         return self._name
 
-    def config_reader(self):
+    def config_reader(self) -> SectionConstraint[SubmoduleConfigParser]:
         """
         :return: ConfigReader instance which allows you to qurey the configuration values
             of this submodule, as provided by the .gitmodules file
@@ -1163,7 +1215,7 @@ class Submodule(IndexObject, Iterable, Traversable):
         :raise IOError: If the .gitmodules file/blob could not be read"""
         return self._config_parser_constrained(read_only=True)
 
-    def children(self):
+    def children(self) -> IterableList['Submodule']:
         """
         :return: IterableList(Submodule, ...) an iterable list of submodules instances
             which are children of this submodule or 0 if the submodule is not checked out"""
@@ -1174,17 +1226,16 @@ class Submodule(IndexObject, Iterable, Traversable):
     #{ Iterable Interface
 
     @classmethod
-    def iter_items(cls, repo, parent_commit='HEAD'):
+    def iter_items(cls, repo: 'Repo', parent_commit: Union[Commit_ish, str] = 'HEAD', *Args: Any, **kwargs: Any
+                   ) -> Iterator['Submodule']:
         """:return: iterator yielding Submodule instances available in the given repository"""
         try:
             pc = repo.commit(parent_commit)         # parent commit instance
             parser = cls._config_parser(repo, pc, read_only=True)
         except (IOError, BadName):
-            return
+            return iter([])
         # END handle empty iterator
 
-        rt = pc.tree                                # root tree
-
         for sms in parser.sections():
             n = sm_name(sms)
             p = parser.get(sms, 'path')
@@ -1197,6 +1248,7 @@ class Submodule(IndexObject, Iterable, Traversable):
             # get the binsha
             index = repo.index
             try:
+                rt = pc.tree                                # root tree
                 sm = rt[p]
             except KeyError:
                 # try the index, maybe it was just added
